@@ -1,10 +1,13 @@
+import uuid
+
 from tooz import coordination
 
-coordinator = coordination.get_coordinator('zookeeper://localhost', b'host-1')
+coordinator = coordination.get_coordinator('zake://', b'host-1')
 coordinator.start()
 
 # Create a group
-request = coordinator.create_group(b"my group")
+group = bytes(str(uuid.uuid4()).encode('ascii'))
+request = coordinator.create_group(group)
 request.get()
 
 
@@ -13,6 +16,5 @@ def group_joined(event):
     print(event.group_id, event.member_id)
 
 
-coordinator.watch_join_group(b"my group", group_joined)
-
+coordinator.watch_join_group(group, group_joined)
 coordinator.stop()
