@@ -12,12 +12,14 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+import functools
 import logging
+import threading
 
+import fasteners
 from oslo_utils import encodeutils
 from oslo_utils import timeutils
 import requests
-import six
 
 import tooz
 from tooz import coordination
@@ -31,19 +33,19 @@ LOG = logging.getLogger(__name__)
 def _translate_failures(func):
     """Translates common requests exceptions into tooz exceptions."""
 
-    @six.wraps(func)
+    @functools.wraps(func)
     def wrapper(*args, **kwargs):
         try:
             return func(*args, **kwargs)
         except ValueError as e:
             # Typically json decoding failed for some reason.
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         except requests.exceptions.RequestException as e:
-            coordination.raise_with_cause(coordination.ToozConnectionError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.ToozConnectionError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
 
     return wrapper
 
@@ -89,17 +91,25 @@ class EtcdLock(locking.Lock):
         super(EtcdLock, self).__init__(name)
         self.client = client
         self.coord = coord
-        self.lock = None
         self.ttl = ttl
         self._lock_url = lock_url
         self._node = None
 
+        # NOTE(jschwarz): this lock is mainly used to prevent concurrent runs
+        # of hearthbeat() with another function. For more details, see
+        # https://bugs.launchpad.net/python-tooz/+bug/1603005.
+        self._lock = threading.Lock()
+
     @_translate_failures
+    @fasteners.locked
     def break_(self):
         reply = self.client.delete(self._lock_url, make_url=False)
         return reply.get('errorCode') is None
 
-    def acquire(self, blocking=True):
+    def acquire(self, blocking=True, shared=False):
+        if shared:
+            raise tooz.NotImplemented
+
         blocking, timeout = utils.convert_blocking(blocking)
         if timeout is not None:
             watch = timeutils.StopWatch(duration=timeout)
@@ -108,22 +118,31 @@ class EtcdLock(locking.Lock):
             watch = None
 
         while True:
-            try:
-                reply = self.client.put(
-                    self._lock_url,
-                    make_url=False,
-                    timeout=watch.leftover() if watch else None,
-                    data={"ttl": self.ttl,
-                          "prevExist": "false"})
-            except requests.exceptions.RequestException:
-                if watch and watch.leftover() == 0:
-                    return False
+            if self.acquired:
+                # We already acquired the lock. Just go ahead and wait for ever
+                # if blocking != False using the last index.
+                lastindex = self._node['modifiedIndex']
+            else:
+                try:
+                    reply = self.client.put(
+                        self._lock_url,
+                        make_url=False,
+                        timeout=watch.leftover() if watch else None,
+                        data={"ttl": self.ttl,
+                              "prevExist": "false"})
+                except requests.exceptions.RequestException:
+                    if not watch or watch.leftover() == 0:
+                        return False
 
-            # We got the lock!
-            if reply.get("errorCode") is None:
-                self._node = reply['node']
-                self.coord._acquired_locks.append(self)
-                return True
+                # We got the lock!
+                if reply.get("errorCode") is None:
+                    with self._lock:
+                        self._node = reply['node']
+                        self.coord._acquired_locks.add(self)
+                    return True
+
+                # No lock, somebody got it, wait for it to be released
+                lastindex = reply['index'] + 1
 
             # We didn't get the lock and we don't want to wait
             if not blocking:
@@ -133,22 +152,23 @@ class EtcdLock(locking.Lock):
             try:
                 reply = self.client.get(
                     self._lock_url +
-                    "?wait=true&waitIndex=%d" % reply['index'],
+                    "?wait=true&waitIndex=%d" % lastindex,
                     make_url=False,
                     timeout=watch.leftover() if watch else None)
             except requests.exceptions.RequestException:
-                if watch and watch.expired():
+                if not watch or watch.expired():
                     return False
 
     @_translate_failures
+    @fasteners.locked
     def release(self):
-        if self in self.coord._acquired_locks:
+        if self.acquired:
             lock_url = self._lock_url
             lock_url += "?prevIndex=%s" % self._node['modifiedIndex']
             reply = self.client.delete(lock_url, make_url=False)
             errorcode = reply.get("errorCode")
             if errorcode is None:
-                self.coord._acquired_locks.remove(self)
+                self.coord._acquired_locks.discard(self)
                 self._node = None
                 return True
             else:
@@ -156,17 +176,26 @@ class EtcdLock(locking.Lock):
                             self.name, errorcode, reply.get('message'))
         return False
 
+    @property
+    def acquired(self):
+        return self in self.coord._acquired_locks
+
     @_translate_failures
+    @fasteners.locked
     def heartbeat(self):
         """Keep the lock alive."""
-        poked = self.client.put(self._lock_url,
-                                data={"ttl": self.ttl,
-                                      "prevExist": "true"}, make_url=False)
-        errorcode = poked.get("errorCode")
-        if errorcode:
+        if self.acquired:
+            poked = self.client.put(self._lock_url,
+                                    data={"ttl": self.ttl,
+                                          "prevExist": "true"}, make_url=False)
+            self._node = poked['node']
+            errorcode = poked.get("errorCode")
+            if not errorcode:
+                return True
             LOG.warning("Unable to heartbeat by updating key '%s' with "
                         "extended expiry of %s seconds: %d, %s", self.name,
                         self.ttl, errorcode, poked.get("message"))
+        return False
 
 
 class EtcdDriver(coordination.CoordinationDriver):
@@ -174,6 +203,21 @@ class EtcdDriver(coordination.CoordinationDriver):
 
     This driver uses etcd provide the coordination driver semantics and
     required API(s).
+
+    The Etcd driver connection URI should look like::
+
+      etcd://[HOST[:PORT]][?OPTION1=VALUE1[&OPTION2=VALUE2[&...]]]
+
+    If not specified, HOST defaults to localhost and PORT defaults to 2379.
+    Available options are:
+
+    ==================  =======
+    Name                Default
+    ==================  =======
+    protocol            http
+    timeout             30
+    lock_timeout        30
+    ==================  =======
     """
 
     #: Default socket/lock/member/leader timeout used when none is provided.
@@ -189,7 +233,7 @@ class EtcdDriver(coordination.CoordinationDriver):
     lock_encoder_cls = utils.Base64LockEncoder
 
     def __init__(self, member_id, parsed_url, options):
-        super(EtcdDriver, self).__init__()
+        super(EtcdDriver, self).__init__(member_id, parsed_url, options)
         host = parsed_url.hostname or self.DEFAULT_HOST
         port = parsed_url.port or self.DEFAULT_PORT
         options = utils.collapse(options)
@@ -198,7 +242,7 @@ class EtcdDriver(coordination.CoordinationDriver):
         default_timeout = options.get('timeout', self.DEFAULT_TIMEOUT)
         self.lock_encoder = self.lock_encoder_cls(self.client.get_url("keys"))
         self.lock_timeout = int(options.get('lock_timeout', default_timeout))
-        self._acquired_locks = []
+        self._acquired_locks = set()
 
     def _start(self):
         try:
@@ -212,30 +256,18 @@ class EtcdDriver(coordination.CoordinationDriver):
                         self, self.client, self.lock_timeout)
 
     def heartbeat(self):
-        for lock in self._acquired_locks:
+        for lock in self._acquired_locks.copy():
             lock.heartbeat()
         return self.lock_timeout
 
-    @staticmethod
-    def watch_join_group(group_id, callback):
+    def watch_join_group(self, group_id, callback):
         raise tooz.NotImplemented
 
-    @staticmethod
-    def unwatch_join_group(group_id, callback):
+    def unwatch_join_group(self, group_id, callback):
         raise tooz.NotImplemented
 
-    @staticmethod
-    def watch_leave_group(group_id, callback):
+    def watch_leave_group(self, group_id, callback):
         raise tooz.NotImplemented
 
-    @staticmethod
-    def unwatch_leave_group(group_id, callback):
-        raise tooz.NotImplemented
-
-    @staticmethod
-    def watch_elected_as_leader(group_id, callback):
-        raise tooz.NotImplemented
-
-    @staticmethod
-    def unwatch_elected_as_leader(group_id, callback):
+    def unwatch_leave_group(self, group_id, callback):
         raise tooz.NotImplemented
