@@ -17,19 +17,21 @@
 import contextlib
 import datetime
 import errno
+import functools
 import hashlib
 import logging
 import os
+import re
 import shutil
+import sys
+import tempfile
 import threading
 import weakref
 
-from concurrent import futures
-
 import fasteners
 from oslo_utils import encodeutils
+from oslo_utils import fileutils
 from oslo_utils import timeutils
-import six
 import voluptuous
 
 import tooz
@@ -44,6 +46,8 @@ class _Barrier(object):
     def __init__(self):
         self.cond = threading.Condition()
         self.owner = None
+        self.shared = False
+        self.ref = 0
 
 
 @contextlib.contextmanager
@@ -51,26 +55,9 @@ def _translate_failures():
     try:
         yield
     except (EnvironmentError, voluptuous.Invalid) as e:
-        coordination.raise_with_cause(coordination.ToozError,
-                                      encodeutils.exception_to_unicode(e),
-                                      cause=e)
-
-
-_SCHEMAS = {
-    'group': voluptuous.Schema({
-        voluptuous.Required('group_id'): voluptuous.Any(six.text_type,
-                                                        six.binary_type),
-        # NOTE(sileht): tooz <1.36 was creating file without this
-        voluptuous.Optional('encoded'): bool,
-    }),
-    'member': voluptuous.Schema({
-        voluptuous.Required('member_id'): voluptuous.Any(six.text_type,
-                                                         six.binary_type),
-        voluptuous.Required('joined_on'): datetime.datetime,
-        # NOTE(sileht): tooz <1.36 was creating file without this
-        voluptuous.Optional('encoded'): bool,
-    }, extra=voluptuous.ALLOW_EXTRA),
-}
+        utils.raise_with_cause(tooz.ToozError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
 
 
 def _convert_from_old_format(data):
@@ -86,32 +73,24 @@ def _convert_from_old_format(data):
     # example of potential old python3 payload:
     # {u"member_id": b"member"}
     # {u"member_id": u"member"}
-    if six.PY3 and b"member_id" in data or b"group_id" in data:
+    if b"member_id" in data or b"group_id" in data:
         data = dict((k.decode("utf8"), v) for k, v in data.items())
         # About member_id and group_id valuse if the file have been written
         # with python2 and in the old format, we can't known with python3
         # if we need to decode the value or not. Python3 see bytes blob
         # We keep it as-is and pray, this have a good change to break if
         # the application was using str in python2 and unicode in python3
-        # The member file is often overriden so it's should be fine
+        # The member file is often overridden so it's should be fine
         # But the group file can be very old, so we
         # now have to update it each time create_group is called
     return data
 
 
-def _load_and_validate(blob, schema_key):
-    data = utils.loads(blob)
-    data = _convert_from_old_format(data)
-    schema = _SCHEMAS[schema_key]
-    schema(data)
-    return data
-
-
 def _lock_me(lock):
 
     def wrapper(func):
 
-        @six.wraps(func)
+        @functools.wraps(func)
         def decorator(*args, **kwargs):
             with lock:
                 return func(*args, **kwargs)
@@ -130,11 +109,12 @@ class FileLock(locking.Lock):
         self._lock = fasteners.InterProcessLock(path)
         self._barrier = barrier
         self._member_id = member_id
+        self.ref = 0
 
     def is_still_owner(self):
         return self.acquired
 
-    def acquire(self, blocking=True):
+    def acquire(self, blocking=True, shared=False):
         blocking, timeout = utils.convert_blocking(blocking)
         watch = timeutils.StopWatch(duration=timeout)
         watch.start()
@@ -142,11 +122,16 @@ class FileLock(locking.Lock):
         # Make the shared barrier ours first.
         with self._barrier.cond:
             while self._barrier.owner is not None:
+                if (shared and self._barrier.shared):
+                    break
                 if not blocking or watch.expired():
                     return False
                 self._barrier.cond.wait(watch.leftover(return_none=True))
             self._barrier.owner = (threading.current_thread().ident,
                                    os.getpid(), self._member_id)
+            self._barrier.shared = shared
+            self._barrier.ref += 1
+            self.ref += 1
 
         # Ok at this point we are now working in a thread safe manner,
         # and now we can try to get the actual lock...
@@ -165,6 +150,8 @@ class FileLock(locking.Lock):
                 # Release the barrier to let someone else have a go at it...
                 with self._barrier.cond:
                     self._barrier.owner = None
+                    self._barrier.ref = 0
+                    self._barrier.shared = False
                     self._barrier.cond.notify_all()
 
         self.acquired = gotten
@@ -174,9 +161,14 @@ class FileLock(locking.Lock):
         if not self.acquired:
             return False
         with self._barrier.cond:
-            self.acquired = False
-            self._barrier.owner = None
-            self._barrier.cond.notify_all()
+            self._barrier.ref -= 1
+            self.ref -= 1
+            if not self.ref:
+                self.acquired = False
+            if not self._barrier.ref:
+                self._barrier.owner = None
+                self._lock.release()
+                self._barrier.cond.notify_all()
         return True
 
     def __del__(self):
@@ -184,8 +176,8 @@ class FileLock(locking.Lock):
             LOG.warning("Unreleased lock %s garbage collected", self.name)
 
 
-class FileDriver(coordination._RunWatchersMixin,
-                 coordination.CoordinationDriver):
+class FileDriver(coordination.CoordinationDriverCachedRunWatchers,
+                 coordination.CoordinationDriverWithExecutor):
     """A file based driver.
 
     This driver uses files and directories (and associated file locks) to
@@ -193,6 +185,13 @@ class FileDriver(coordination._RunWatchersMixin,
     missing some functionality but in the future these not implemented API(s)
     will be filled in.
 
+    The File driver connection URI should look like::
+
+      file://DIRECTORY[?timeout=TIMEOUT]
+
+    DIRECTORY is the location that should be used to store lock files.
+    TIMEOUT defaults to 10.
+
     General recommendations/usage considerations:
 
     - It does **not** automatically delete members from
@@ -227,19 +226,31 @@ class FileDriver(coordination._RunWatchersMixin,
 
     def __init__(self, member_id, parsed_url, options):
         """Initialize the file driver."""
-        super(FileDriver, self).__init__()
-        self._member_id = member_id
-        self._dir = parsed_url.path
-        self._executor = utils.ProxyExecutor.build("File", options)
+        super(FileDriver, self).__init__(member_id, parsed_url, options)
+        self._dir = self._normalize_path(parsed_url.path)
         self._group_dir = os.path.join(self._dir, 'groups')
+        self._tmpdir = os.path.join(self._dir, 'tmp')
         self._driver_lock_path = os.path.join(self._dir, '.driver_lock')
         self._driver_lock = self._get_raw_lock(self._driver_lock_path,
                                                self._member_id)
-        self._reserved_dirs = [self._dir, self._group_dir]
+        self._reserved_dirs = [self._dir, self._group_dir, self._tmpdir]
         self._reserved_paths = list(self._reserved_dirs)
         self._reserved_paths.append(self._driver_lock_path)
-        self._joined_groups = set()
         self._safe_member_id = self._make_filesystem_safe(member_id)
+        self._timeout = int(self._options.get('timeout', 10))
+
+    @staticmethod
+    def _normalize_path(path):
+        if sys.platform == 'win32':
+            # Replace slashes with backslashes and make sure we don't
+            # have any at the beginning of paths that include drive letters.
+            #
+            # Expected url format:
+            # file:////share_address/share_name
+            # file:///C:/path
+            return re.sub(r'\\(?=\w:\\)', '',
+                          os.path.normpath(path))
+        return path
 
     @classmethod
     def _get_raw_lock(cls, path, member_id):
@@ -260,17 +271,12 @@ class FileDriver(coordination._RunWatchersMixin,
         return hashlib.new(cls.HASH_ROUTINE, item).hexdigest()
 
     def _start(self):
+        super(FileDriver, self)._start()
         for a_dir in self._reserved_dirs:
             try:
-                utils.ensure_tree(a_dir)
+                fileutils.ensure_tree(a_dir)
             except OSError as e:
                 raise coordination.ToozConnectionError(e)
-        self._executor.start()
-
-    def _stop(self):
-        while self._joined_groups:
-            self.leave_group(self._joined_groups.pop())
-        self._executor.stop()
 
     def _update_group_metadata(self, path, group_id):
         details = {
@@ -278,17 +284,18 @@ class FileDriver(coordination._RunWatchersMixin,
         }
         details[u'encoded'] = details[u"group_id"] != group_id
         details_blob = utils.dumps(details)
-        with open(path, "wb") as fh:
+        fd, name = tempfile.mkstemp("tooz", dir=self._tmpdir)
+        with os.fdopen(fd, "wb") as fh:
             fh.write(details_blob)
+        os.rename(name, path)
 
     def create_group(self, group_id):
         safe_group_id = self._make_filesystem_safe(group_id)
         group_dir = os.path.join(self._group_dir, safe_group_id)
         group_meta_path = os.path.join(group_dir, '.metadata')
 
-        @_lock_me(self._driver_lock)
         def _do_create_group():
-            if os.path.isdir(group_dir):
+            if os.path.exists(os.path.join(group_dir, ".metadata")):
                 # NOTE(sileht): We update the group metadata even
                 # they are already good, so ensure dict key are convert
                 # to unicode in case of the file have been written with
@@ -296,7 +303,7 @@ class FileDriver(coordination._RunWatchersMixin,
                 self._update_group_metadata(group_meta_path, group_id)
                 raise coordination.GroupAlreadyExist(group_id)
             else:
-                utils.ensure_tree(group_dir)
+                fileutils.ensure_tree(group_dir)
                 self._update_group_metadata(group_meta_path, group_id)
         fut = self._executor.submit(_do_create_group)
         return FileFutureResult(fut)
@@ -308,7 +315,7 @@ class FileDriver(coordination._RunWatchersMixin,
 
         @_lock_me(self._driver_lock)
         def _do_join_group():
-            if not os.path.isdir(group_dir):
+            if not os.path.exists(os.path.join(group_dir, ".metadata")):
                 raise coordination.GroupNotCreated(group_id)
             if os.path.isfile(me_path):
                 raise coordination.MemberAlreadyExist(group_id,
@@ -335,7 +342,7 @@ class FileDriver(coordination._RunWatchersMixin,
 
         @_lock_me(self._driver_lock)
         def _do_leave_group():
-            if not os.path.isdir(group_dir):
+            if not os.path.exists(os.path.join(group_dir, ".metadata")):
                 raise coordination.GroupNotCreated(group_id)
             try:
                 os.unlink(me_path)
@@ -351,23 +358,42 @@ class FileDriver(coordination._RunWatchersMixin,
         fut = self._executor.submit(_do_leave_group)
         return FileFutureResult(fut)
 
+    _SCHEMAS = {
+        'group': voluptuous.Schema({
+            voluptuous.Required('group_id'): voluptuous.Any(str, bytes),
+            # NOTE(sileht): tooz <1.36 was creating file without this
+            voluptuous.Optional('encoded'): bool,
+        }),
+        'member': voluptuous.Schema({
+            voluptuous.Required('member_id'): voluptuous.Any(str, bytes),
+            voluptuous.Required('joined_on'): datetime.datetime,
+            # NOTE(sileht): tooz <1.36 was creating file without this
+            voluptuous.Optional('encoded'): bool,
+        }, extra=voluptuous.ALLOW_EXTRA),
+    }
+
+    def _load_and_validate(self, blob, schema_key):
+        data = utils.loads(blob)
+        data = _convert_from_old_format(data)
+        schema = self._SCHEMAS[schema_key]
+        return schema(data)
+
+    def _read_member_id(self, path):
+        with open(path, 'rb') as fh:
+            details = self._load_and_validate(fh.read(), 'member')
+            if details.get("encoded"):
+                return details[u'member_id'].decode("utf-8")
+            return details[u'member_id']
+
     def get_members(self, group_id):
         safe_group_id = self._make_filesystem_safe(group_id)
         group_dir = os.path.join(self._group_dir, safe_group_id)
 
-        def _read_member_id(path):
-            with open(path, 'rb') as fh:
-                details = _load_and_validate(fh.read(), 'member')
-                if details.get("encoded"):
-                    return details[u'member_id'].decode("utf-8")
-                else:
-                    return details[u'member_id']
-
         @_lock_me(self._driver_lock)
         def _do_get_members():
             if not os.path.isdir(group_dir):
                 raise coordination.GroupNotCreated(group_id)
-            members = []
+            members = set()
             try:
                 entries = os.listdir(group_dir)
             except EnvironmentError as e:
@@ -380,12 +406,20 @@ class FileDriver(coordination._RunWatchersMixin,
                         continue
                     entry_path = os.path.join(group_dir, entry)
                     try:
-                        member_id = _read_member_id(entry_path)
+                        m_time = datetime.datetime.fromtimestamp(
+                            os.stat(entry_path).st_mtime)
+                        current_time = datetime.datetime.now()
+                        delta_time = timeutils.delta_seconds(m_time,
+                                                             current_time)
+                        if delta_time >= 0 and delta_time <= self._timeout:
+                            member_id = self._read_member_id(entry_path)
+                        else:
+                            continue
                     except EnvironmentError as e:
                         if e.errno != errno.ENOENT:
                             raise
                     else:
-                        members.append(member_id)
+                        members.add(member_id)
             return members
 
         fut = self._executor.submit(_do_get_members)
@@ -412,7 +446,7 @@ class FileDriver(coordination._RunWatchersMixin,
                 else:
                     raise
             else:
-                details = _load_and_validate(contents, 'member')
+                details = self._load_and_validate(contents, 'member')
                 return details.get(u"capabilities")
 
         fut = self._executor.submit(_do_get_member_capabilities)
@@ -434,6 +468,11 @@ class FileDriver(coordination._RunWatchersMixin,
             else:
                 if len(entries) > 1:
                     raise coordination.GroupNotEmpty(group_id)
+                elif len(entries) == 1 and entries != ['.metadata']:
+                    raise tooz.ToozError(
+                        "Unexpected path '%s' found in"
+                        " group directory '%s' (expected to only find"
+                        " a '.metadata' path)" % (entries[0], group_dir))
                 else:
                     try:
                         shutil.rmtree(group_dir)
@@ -444,23 +483,21 @@ class FileDriver(coordination._RunWatchersMixin,
         fut = self._executor.submit(_do_delete_group)
         return FileFutureResult(fut)
 
+    def _read_group_id(self, path):
+        with open(path, 'rb') as fh:
+            details = self._load_and_validate(fh.read(), 'group')
+            if details.get("encoded"):
+                return details[u'group_id'].decode("utf-8")
+            return details[u'group_id']
+
     def get_groups(self):
 
-        def _read_group_id(path):
-            with open(path, 'rb') as fh:
-                details = _load_and_validate(fh.read(), 'group')
-                if details.get("encoded"):
-                    return details[u'group_id'].decode("utf-8")
-                else:
-                    return details[u'group_id']
-
-        @_lock_me(self._driver_lock)
         def _do_get_groups():
             groups = []
             for entry in os.listdir(self._group_dir):
                 path = os.path.join(self._group_dir, entry, '.metadata')
                 try:
-                    groups.append(_read_group_id(path))
+                    groups.append(self._read_group_id(path))
                 except EnvironmentError as e:
                     if e.errno != errno.ENOENT:
                         raise
@@ -469,24 +506,22 @@ class FileDriver(coordination._RunWatchersMixin,
         fut = self._executor.submit(_do_get_groups)
         return FileFutureResult(fut)
 
-    def _init_watch_group(self, group_id):
-        group_members_fut = self.get_members(group_id)
-        group_members = group_members_fut.get(timeout=None)
-        self._group_members[group_id].update(group_members)
+    def heartbeat(self):
+        for group_id in self._joined_groups:
+            safe_group_id = self._make_filesystem_safe(group_id)
+            group_dir = os.path.join(self._group_dir, safe_group_id)
+            member_path = os.path.join(group_dir, "%s.raw" %
+                                       self._safe_member_id)
 
-    def watch_join_group(self, group_id, callback):
-        self._init_watch_group(group_id)
-        return super(FileDriver, self).watch_join_group(group_id, callback)
-
-    def unwatch_join_group(self, group_id, callback):
-        return super(FileDriver, self).unwatch_join_group(group_id, callback)
-
-    def watch_leave_group(self, group_id, callback):
-        self._init_watch_group(group_id)
-        return super(FileDriver, self).watch_leave_group(group_id, callback)
-
-    def unwatch_leave_group(self, group_id, callback):
-        return super(FileDriver, self).unwatch_leave_group(group_id, callback)
+            @_lock_me(self._driver_lock)
+            def _do_heartbeat():
+                try:
+                    os.utime(member_path, None)
+                except EnvironmentError as err:
+                    if err.errno != errno.ENOENT:
+                        raise
+            _do_heartbeat()
+        return self._timeout
 
     @staticmethod
     def watch_elected_as_leader(group_id, callback):
@@ -497,23 +532,5 @@ class FileDriver(coordination._RunWatchersMixin,
         raise tooz.NotImplemented
 
 
-class FileFutureResult(coordination.CoordAsyncResult):
-    """File asynchronous result that references a future."""
-
-    def __init__(self, fut):
-        self._fut = fut
-
-    def get(self, timeout=None):
-        try:
-            # Late translate the common failures since the file driver
-            # may throw things that we can not catch in the callbacks where
-            # it is used.
-            with _translate_failures():
-                return self._fut.result(timeout=timeout)
-        except futures.TimeoutError as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
-
-    def done(self):
-        return self._fut.done()
+FileFutureResult = functools.partial(coordination.CoordinatorResult,
+                                     failure_translator=_translate_failures)
