@@ -12,8 +12,6 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
-from __future__ import absolute_import
-
 from zake import fake_client
 from zake import fake_storage
 
@@ -28,11 +26,12 @@ class ZakeDriver(zookeeper.KazooDriver):
     this manner) for testing and integration (where real `zookeeper`_ servers
     are typically not available).
 
-    .. _zake: https://pypi.python.org/pypi/zake
+    .. _zake: https://pypi.org/project/zake
     .. _zookeeper: http://zookeeper.apache.org/
     """
 
     CHARACTERISTICS = (
+        coordination.Characteristics.NON_TIMEOUT_BASED,
         coordination.Characteristics.DISTRIBUTED_ACROSS_THREADS,
     )
     """
