@@ -16,7 +16,6 @@
 
 import base64
 import datetime
-import errno
 import operator
 import os
 
@@ -24,9 +23,9 @@ import futurist
 import msgpack
 from oslo_serialization import msgpackutils
 from oslo_utils import encodeutils
-import six
+from oslo_utils import excutils
 
-from tooz import coordination
+import tooz
 
 
 class Base64LockEncoder(object):
@@ -36,7 +35,7 @@ class Base64LockEncoder(object):
             self.keyspace_url += prefix
 
     def check_and_encode(self, name):
-        if not isinstance(name, (six.text_type, six.binary_type)):
+        if not isinstance(name, (str, bytes)):
             raise TypeError("Provided lock name is expected to be a string"
                             " or binary type and not %s" % type(name))
         try:
@@ -47,7 +46,7 @@ class Base64LockEncoder(object):
                              % encodeutils.exception_to_unicode(e))
 
     def encode(self, name):
-        if isinstance(name, six.text_type):
+        if isinstance(name, str):
             name = name.encode("ascii")
         enc_name = base64.urlsafe_b64encode(name)
         return self.keyspace_url + "/" + enc_name.decode("ascii")
@@ -83,10 +82,10 @@ class ProxyExecutor(object):
                 default_executor_fact = cls.KIND_TO_FACTORY[executor_kind]
             except KeyError:
                 executors_known = sorted(list(cls.KIND_TO_FACTORY))
-                raise coordination.ToozError("Unknown executor"
-                                             " '%s' provided, accepted values"
-                                             " are %s" % (executor_kind,
-                                                          executors_known))
+                raise tooz.ToozError("Unknown executor"
+                                     " '%s' provided, accepted values"
+                                     " are %s" % (executor_kind,
+                                                  executors_known))
         return cls(driver_name, default_executor_fact)
 
     def start(self):
@@ -104,14 +103,14 @@ class ProxyExecutor(object):
 
     def submit(self, cb, *args, **kwargs):
         if not self.started:
-            raise coordination.ToozError("%s driver asynchronous executor"
-                                         " has not been started"
-                                         % self.driver_name)
+            raise tooz.ToozError("%s driver asynchronous executor"
+                                 " has not been started"
+                                 % self.driver_name)
         try:
             return self.executor.submit(cb, *args, **kwargs)
         except RuntimeError:
-            raise coordination.ToozError("%s driver asynchronous executor has"
-                                         " been shutdown" % self.driver_name)
+            raise tooz.ToozError("%s driver asynchronous executor has"
+                                 " been shutdown" % self.driver_name)
 
 
 def safe_abs_path(rooted_at, *pieces):
@@ -139,21 +138,6 @@ def convert_blocking(blocking):
     return blocking, timeout
 
 
-def ensure_tree(path):
-    """Create a directory (and any ancestor directories required).
-
-    :param path: Directory to create
-    """
-    try:
-        os.makedirs(path)
-    except EnvironmentError as e:
-        if e.errno != errno.EEXIST or not os.path.isdir(path):
-            raise
-        return False
-    else:
-        return True
-
-
 def collapse(config, exclude=None, item_selector=operator.itemgetter(-1)):
     """Collapses config with keys and **list/tuple** values.
 
@@ -169,7 +153,7 @@ def collapse(config, exclude=None, item_selector=operator.itemgetter(-1)):
     if exclude is None:
         exclude = set()
     collapsed = {}
-    for (k, v) in six.iteritems(config):
+    for (k, v) in config.items():
         if isinstance(v, (tuple, list)):
             if k in exclude:
                 collapsed[k] = v
@@ -183,31 +167,58 @@ def collapse(config, exclude=None, item_selector=operator.itemgetter(-1)):
 
 def to_binary(text, encoding='ascii'):
     """Return the binary representation of string (if not already binary)."""
-    if not isinstance(text, six.binary_type):
+    if not isinstance(text, bytes):
         text = text.encode(encoding)
     return text
 
 
-def dumps(data, excp_cls=coordination.SerializationError):
+class SerializationError(tooz.ToozError):
+    "Exception raised when serialization or deserialization breaks."
+
+
+def dumps(data, excp_cls=SerializationError):
     """Serializes provided data using msgpack into a byte string."""
     try:
         return msgpackutils.dumps(data)
     except (msgpack.PackException, ValueError) as e:
-        coordination.raise_with_cause(excp_cls,
-                                      encodeutils.exception_to_unicode(e),
-                                      cause=e)
+        raise_with_cause(excp_cls,
+                         encodeutils.exception_to_unicode(e),
+                         cause=e)
 
 
-def loads(blob, excp_cls=coordination.SerializationError):
+def loads(blob, excp_cls=SerializationError):
     """Deserializes provided data using msgpack (from a prior byte string)."""
     try:
         return msgpackutils.loads(blob)
     except (msgpack.UnpackException, ValueError) as e:
-        coordination.raise_with_cause(excp_cls,
-                                      encodeutils.exception_to_unicode(e),
-                                      cause=e)
+        raise_with_cause(excp_cls,
+                         encodeutils.exception_to_unicode(e),
+                         cause=e)
 
 
 def millis_to_datetime(milliseconds):
     """Converts number of milliseconds (from epoch) into a datetime object."""
     return datetime.datetime.fromtimestamp(float(milliseconds) / 1000)
+
+
+def raise_with_cause(exc_cls, message, *args, **kwargs):
+    """Helper to raise + chain exceptions (when able) and associate a *cause*.
+
+    **For internal usage only.**
+
+    NOTE(harlowja): Since in py3.x exceptions can be chained (due to
+    :pep:`3134`) we should try to raise the desired exception with the given
+    *cause*.
+
+    :param exc_cls: the :py:class:`~tooz.ToozError` class to raise.
+    :param message: the text/str message that will be passed to
+                    the exceptions constructor as its first positional
+                    argument.
+    :param args: any additional positional arguments to pass to the
+                 exceptions constructor.
+    :param kwargs: any additional keyword arguments to pass to the
+                   exceptions constructor.
+    """
+    if not issubclass(exc_cls, tooz.ToozError):
+        raise ValueError("Subclass of tooz error is required")
+    excutils.raise_with_cause(exc_cls, message, *args, **kwargs)
